# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


from flask import Blueprint

from rains.common.api.common import *


task = Blueprint('task', __name__)


@task.route('/task/test', methods=['GET'])
def test():
    return 'test'



# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


from rains.common.decorator import singleton_pattern


@singleton_pattern
class CaseRecorder(object):
    """
    用例记录器

    * 该模块负责记录用例执行时产生的日志，并且暂存至下一次读取。

    """

    _record_dict: dict
    """ 记录列表字典 """

    def __init__(self):
        """
        初始化
        """
        self._record_dict = {}

    @property
    def record_dict(self):
        return self._record_dict

    def write(self, core_id, record: str):
        """
        录入
        """
        if core_id in self.record_dict:
            self.record_dict.update({core_id: ''.join([self.record_dict[core_id], '\n', record])})

        else:
            self.record_dict[core_id] = record

    def read(self, core_id) -> str:
        """
        读取

        * 读取后，缓冲器会清空所有暂存数据。

        Return:
            所有记录拼接成的字符串

        """
        return_record_str = self._record_dict[core_id]

        del self._record_dict[core_id]

        return return_record_str




from rains.kit.web import *
from rains.common.run_pool import RunPool


class TaskTest(WebTask):
    """
    测试任务
    """
    user_input: WebElement
    password_input: WebElement
    login_button: WebElement
    create_button: WebElement

    def set_class_starting(self):
        self.user_input = self.plant.element(
            page='超管后台',
            name='邮箱输入框',
            by_key=BY.XPATH,
            by_value='/html/body/div/div/div/div/div[2]/form/div[1]/div/input'
        )

        self.password_input = self.plant.element(
            page='超管后台',
            name='密码输入框',
            by_key=BY.XPATH,
            by_value='/html/body/div/div/div/div/div[2]/form/div[2]/div/input'
        )

        self.login_button = self.plant.element(
            page='超管后台',
            name='Login按钮',
            by_key=BY.XPATH,
            by_value='/html/body/div/div/div/div/div[2]/form/div[4]/div/button'
        )

        self.create_button = self.plant.element(
            page='超管后台',
            name='Login按钮',
            by_key=BY.XPATH,
            by_value='/html/body/div/div/div/div/div/div[1]/a'
        )

    def set_function_starting(self):
        self.plant.view.to_url('http://manage.dev-tea.cblink.net/')

    def case_01(self):
        """
        登录失败
        """
        self.user_input.input.send('admin@baocai.us')
        self.password_input.input.send('654321')
        self.login_button.mouse.tap()
        return self.create_button.get.in_page() is not True

    def case_02(self):
        """
        登录成功
        """
        self.user_input.input.send('admin@baocai.us')
        self.password_input.input.send('123456')
        self.login_button.mouse.tap()
        return self.create_button.get.in_page() is True


if __name__ == '__main__':

    # core = WebCore()
    # core.start_task(TaskTest)

    tasks = []
    for i in range(20):
        tasks.append(TaskTest)

    RunPool(
        cores=[WebCore(), WebCore(), WebCore()],
        tasks=tasks
    ).running()



# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


from abc import ABCMeta
from abc import abstractmethod


class IEngine(metaclass=ABCMeta):
    """
    引擎接口
    """

    @abstractmethod
    def get_state(self) -> bool:
        """
        获取引擎状态

        * 获取当前引擎状态，启动中返回 True，否则返回 False。

        Return:
            当前引擎状态

        """

    @abstractmethod
    def run(self):
        """
        运行引擎

        * 驱动与定位器的实例化。

        """

    @abstractmethod
    def end(self):
        """
        结束引擎

        * 退出并销毁当前持有的驱动与定位器。

        """

    @abstractmethod
    def reset(self):
        """
        重置引擎

        * 将引擎重置至初始阶段。
        * 与 end() 相比，该函数只是重置引擎至初始状态，而不是注销。

        """



# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


import random

from rains.common.i.i_task import ITask
from rains.common.const import ConstTaskType
from rains.kit.web.plant import WebPlant


class WebTask(ITask):
    """
    Web任务集
    """

    _task_id: int = random.randint(10000, 99999)
    """ 任务编号 """

    _type: str = ConstTaskType.CODE
    """ 任务类型 """

    plant: WebPlant
    """ 工厂对象 """

    def set_class_starting(self):
        pass

    def set_class_ending(self):
        pass

    def set_function_starting(self):
        pass

    def set_function_ending(self):
        pass



# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


from abc import ABCMeta
from abc import abstractmethod


class ITask(metaclass=ABCMeta):
    """
    任务接口

    * 任务集以函数为单位描述执行用例。
    * 运行路径优先级::
    * set_class_starting(类起点)
    * set_function_starting(函数起点)
    * case_xxx(用例函数)
    * set_function_ending(函数终点)
    * set_class_ending(类终点)

    """

    @abstractmethod
    def set_class_starting(self):
        """
        设置类起点

        * 该接口将在 [ 任务类 ] 开始后执行，全程只会执行一次。

        """

    @abstractmethod
    def set_class_ending(self):
        """
        设置类终点

        * 该接口将在 [ 任务类 ] 结束后执行，全程只会执行一次。

        """

    @abstractmethod
    def set_function_starting(self):
        """
        设置函数起点

        * 该接口将在每次 [ 任务类::函数 ] 开始前执行。

        """

    @abstractmethod
    def set_function_ending(self):
        """
        设置函数终点

        * 该接口将在每次 [ 任务类::函数 ] 结束后执行。

        """



# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


def singleton_pattern(cls, *args, **kwargs):
    """
    单例模式装饰器
    """

    instances = {}

    def _singleton_pattern():
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]

    return _singleton_pattern



# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


from rains.kit.web.task import WebTask
from rains.kit.web.core import WebCore
from rains.kit.web.core import BROWSER_TYPE
from rains.kit.web.plant import BY
from rains.kit.web.plant import WebPlant
from rains.kit.web.plant import WebPlantElement
from rains.kit.web.model import WebModel

WebElement = WebPlantElement

__all__ = ['WebTask', 'WebCore', 'BROWSER_TYPE', 'BY', 'WebPlant', 'WebElement', 'WebModel']



# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


from rains.common.const import ConstTaskAndCaseState
from rains.common.db.const import ConstDatabaseTaskNaming
from rains.common.db.const import ConstDatabaseCaseNaming
from rains.common.db.const import ConstDatabaseTableNaming


class Sql(object):
    """
    SQL语句

    * 动态拼接 SQLite 语句。

    """

    @staticmethod
    def tasks_get_count() -> str:
        """
        查询所有任务数量
        """
        return f""" 

        SELECT COUNT(id)

        FROM { ConstDatabaseTableNaming.TASKS }

        """

    @staticmethod
    def tasks_get_count_in_fail() -> str:
        """
        查询所有任务数量
        """
        return f""" 

        SELECT COUNT(id)

        FROM { ConstDatabaseTableNaming.TASKS }
        
        WHERE { ConstDatabaseTaskNaming.CASE_FAIL } != 0

        """

    @staticmethod
    def tasks_get_count_from_data(paras: dict = None) -> str:
        """
        查询指定日期中的所有任务数量

        参数:
            * execute (str): 执行日期，不可为空

        """
        # 处理参数
        paras = _machining_parameter(paras, [ConstDatabaseTaskNaming.EXECUTE_DATE])

        return f""" 

        SELECT COUNT(id)

        FROM { ConstDatabaseTableNaming.TASKS }
        
        WHERE { ConstDatabaseTaskNaming.EXECUTE_DATE } = '{ paras[ConstDatabaseTaskNaming.EXECUTE_DATE] }'

        """

    @staticmethod
    def tasks_get_spend_time() -> str:
        """
        查询所有任务的消耗时间
        """
        return f""" 

        SELECT SUM({ ConstDatabaseTaskNaming.SPEND_TIME_S })

        FROM { ConstDatabaseTableNaming.TASKS }

        """

    @staticmethod
    def tasks_get_spend_time_from_data(paras: dict = None) -> str:
        """
        查询指定日期中所有任务的消耗时间
        """
        # 处理参数
        paras = _machining_parameter(paras, [ConstDatabaseTaskNaming.EXECUTE_DATE])

        return f""" 

        SELECT SUM({ ConstDatabaseTaskNaming.SPEND_TIME_S })

        FROM { ConstDatabaseTableNaming.TASKS }
        
        WHERE { ConstDatabaseTaskNaming.EXECUTE_DATE } = '{ paras[ConstDatabaseTaskNaming.EXECUTE_DATE] }'

        """

    @staticmethod
    def tasks_get_item(paras: dict = None) -> str:
        """
        查询任务

        * 返回指定任务ID的任务信息。

        参数:
            * tid (int): 任务ID，不可为空

        """
        # 处理参数
        paras = _machining_parameter(paras, [ConstDatabaseCaseNaming.TID])

        return f"""

        SELECT *

        FROM { ConstDatabaseTableNaming.TASKS }

        WHERE ID = { paras[ConstDatabaseCaseNaming.TID] }

        """

    @staticmethod
    def tasks_get_all_item(paras: dict = None) -> str:
        """
        查询所有任务

        * 返回所有任务信息列表。

        参数:
            * page (int): 页数
            * number (int): 查询个数

        """
        # 处理参数
        paras = _machining_parameter(paras, [])
        # 获取数据返回量限制
        limit = _get_desc_limit_section(paras)

        return f"""

        SELECT * 
        
        FROM { ConstDatabaseTableNaming.TASKS }

        ORDER BY ID DESC 

        { limit }

        """

    @staticmethod
    def tasks_get_date_list(paras: dict = None) -> str:
        """
        查询工作日列表

        * 返回去重后的所有包含任务信息的日期列表。

        参数:
            * page (int): 页数
            * number (int): 查询个数

        """
        # 处理参数
        paras = _machining_parameter(paras, [])
        # 获取数据返回量限制
        limit = _get_desc_limit_section(paras)

        return f"""

        SELECT DISTINCT { ConstDatabaseTaskNaming.EXECUTE_DATE } 
        
        FROM { ConstDatabaseTableNaming.TASKS }
        
        ORDER BY ID DESC

        { limit }

        """

    @staticmethod
    def tasks_add(paras: dict = None) -> str:
        """
        创建任务

        * 创建任务执行信息。

        参数:
            * name (str): 任务名称，不可为空
            * remark (str): 任务备注，不可为空
            * execute_date (date): 执行日期，不可为空

        """
        # 处理参数
        paras = _machining_parameter(paras,
                                     [
                                         ConstDatabaseTaskNaming.NAME,
                                         ConstDatabaseTaskNaming.REMARK,
                                         ConstDatabaseTaskNaming.EXECUTE_DATE
                                     ])

        return f""" 

        INSERT INTO { ConstDatabaseTableNaming.TASKS } (
        
             id, 
             { ConstDatabaseTaskNaming.NAME },
             { ConstDatabaseTaskNaming.REMARK },
             { ConstDatabaseTaskNaming.EXECUTE_DATE },
             { ConstDatabaseTaskNaming.IS_COMPLETED },
             { ConstDatabaseTaskNaming.START_TIME },
             { ConstDatabaseTaskNaming.END_TIME },
             { ConstDatabaseTaskNaming.SPEND_TIME_S },
             { ConstDatabaseTaskNaming.CASE_ALL },
             { ConstDatabaseTaskNaming.CASE_PASS },
             { ConstDatabaseTaskNaming.CASE_FAIL }
        )
        
        VALUES (
        
            NULL,
            '{ paras[ConstDatabaseTaskNaming.NAME] }',
            '{ paras[ConstDatabaseTaskNaming.REMARK] }',
            '{ paras[ConstDatabaseTaskNaming.EXECUTE_DATE] }',
             0, NULL, NULL, NULL, NULL, NULL, NULL
        )

        """

    @staticmethod
    def tasks_update(paras: dict = None) -> str:
        """
        更新任务

        * 更新任务信息。

        参数:
            * tid (int): 所属任务ID，不可为空
            * is_completed (bool): 是否已完成，不可为空
            * start_time (date): 开始时间，不可为空
            * end_time (date): 结束时间，不可为空
            * spend_time_s (int): 花费时间(秒)，不可为空
            * case_all (int): 所有用例数，不可为空
            * case_pass (int): 成功的用例数，不可为空
            * case_fail (int): 失败用例数量，不可为空

        """
        # 处理参数
        paras = _machining_parameter(paras,
                                     [
                                         ConstDatabaseTaskNaming.TID,
                                         ConstDatabaseTaskNaming.IS_COMPLETED,
                                         ConstDatabaseTaskNaming.START_TIME,
                                         ConstDatabaseTaskNaming.END_TIME,
                                         ConstDatabaseTaskNaming.SPEND_TIME_S,
                                         ConstDatabaseTaskNaming.CASE_ALL,
                                         ConstDatabaseTaskNaming.CASE_PASS,
                                         ConstDatabaseTaskNaming.CASE_FAIL
                                     ])

        return f"""

        UPDATE { ConstDatabaseTableNaming.TASKS }

        SET
            { ConstDatabaseTaskNaming.IS_COMPLETED }  = { paras[ConstDatabaseTaskNaming.IS_COMPLETED] },
            { ConstDatabaseTaskNaming.START_TIME }    = '{ paras[ConstDatabaseTaskNaming.START_TIME] }',
            { ConstDatabaseTaskNaming.END_TIME }      = '{ paras[ConstDatabaseTaskNaming.END_TIME] }',
            { ConstDatabaseTaskNaming.SPEND_TIME_S }  = '{ paras[ConstDatabaseTaskNaming.SPEND_TIME_S] }',
            { ConstDatabaseTaskNaming.CASE_ALL }      =  { paras[ConstDatabaseTaskNaming.CASE_ALL] },
            { ConstDatabaseTaskNaming.CASE_PASS }     =  { paras[ConstDatabaseTaskNaming.CASE_PASS] },
            { ConstDatabaseTaskNaming.CASE_FAIL }     =  { paras[ConstDatabaseTaskNaming.CASE_FAIL] }

        WHERE ID = { paras[ConstDatabaseTaskNaming.TID] }

        """

    @staticmethod
    def tasks_delete(paras: dict = None) -> str:
        """
        删除任务

        * 删除指定任务ID的任务信息。

        参数:
            * tid (int): 所属任务ID，不可为空

        """
        # 处理参数
        paras = _machining_parameter(paras, [ConstDatabaseTaskNaming.TID])

        return f"""

        DELETE FROM { ConstDatabaseTableNaming.TASKS }

        WHERE ID = { paras[ConstDatabaseTaskNaming.TID] }

        """

    @staticmethod
    def cases_get_all_count() -> str:
        """
        查询所有用例数量
        """
        return f""" 

        SELECT COUNT(id) FROM { ConstDatabaseTableNaming.CASES } 

        """

    @staticmethod
    def cases_get_all_count_in_fail() -> str:
        """
        查询所有失败的用例数量
        """
        return f""" 

        SELECT COUNT(id) FROM { ConstDatabaseTableNaming.CASES }
        
        WHERE { ConstDatabaseCaseNaming.STATE } != '{ ConstTaskAndCaseState.SUCCESSFUL }'

        """

    @staticmethod
    def cases_get_all_count_in_fail_from_data(paras: dict = None) -> str:
        """
        查询指定日期里所有失败的用例数量
        """
        # 处理参数
        paras = _machining_parameter(paras, [ConstDatabaseCaseNaming.EXECUTE_DATE])

        return f""" 

        SELECT COUNT(id) FROM {ConstDatabaseTableNaming.CASES}

        WHERE {ConstDatabaseCaseNaming.STATE} != '{ConstTaskAndCaseState.SUCCESSFUL}'
        
        AND { ConstDatabaseCaseNaming.EXECUTE_DATE } == '{ paras[ConstDatabaseCaseNaming.EXECUTE_DATE] }'

        """

    @staticmethod
    def cases_get_count(paras: dict = None) -> str:
        """
        查询指定任务下所有用例数量

        参数:
            * tid (int): 所属任务ID，不可为空

        """
        # 处理参数
        paras = _machining_parameter(paras, [ConstDatabaseTaskNaming.TID])

        return f""" 

        SELECT COUNT(id) FROM { ConstDatabaseTableNaming.CASES } 
        
        WHERE { ConstDatabaseTaskNaming.TID } = { paras[ConstDatabaseTaskNaming.TID] }

        """

    @staticmethod
    def cases_get_count_form_data(paras: dict = None) -> str:
        """
        查询指定任务下所有用例数量

        参数:
            * execute (str): 执行日期，不可为空

        """
        # 处理参数
        paras = _machining_parameter(paras, [ConstDatabaseTaskNaming.EXECUTE_DATE])

        return f""" 

        SELECT COUNT(id) FROM {ConstDatabaseTableNaming.CASES} 

        WHERE {ConstDatabaseTaskNaming.EXECUTE_DATE} = '{paras[ConstDatabaseTaskNaming.EXECUTE_DATE]}'

        """

    @staticmethod
    def cases_get_item(paras: dict = None) -> str:
        """
        查询用例

        * 返回指定用例ID的用例信息。

        参数:
            * cid (int): 用例ID，不可为空

        """
        # 处理参数
        paras = _machining_parameter(paras, [ConstDatabaseCaseNaming.CID])

        return f"""

        SELECT * FROM { ConstDatabaseTableNaming.CASES }

        WHERE id = { paras[ConstDatabaseCaseNaming.CID] }

        """

    @staticmethod
    def cases_get_all_item(paras: dict = None) -> str:
        """
        查询用例列表

        * 返回所有用例信息列表

        参数:
            * tid (int): 所属任务ID，不可为空
            * state (str): 状态
            * page (int): 页数
            * number (int): 查询个数

        """
        # 处理参数
        paras = _machining_parameter(paras, [ConstDatabaseCaseNaming.TID])

        if ConstDatabaseCaseNaming.STATE in paras.keys():
            state = f'{ ConstDatabaseCaseNaming.STATE } = { paras[ConstDatabaseCaseNaming.STATE] }'
        else:
            state = f'{ ConstDatabaseCaseNaming.STATE } in ("{ ConstTaskAndCaseState.BLOCK }", ' \
                    f'"{ ConstTaskAndCaseState.SUCCESSFUL }", "{ ConstTaskAndCaseState.UNSUCCESSFUL }")'

        # 获取数据返回量限制
        limit = _get_desc_limit_section(paras)

        return f"""

        SELECT * FROM { ConstDatabaseTableNaming.CASES }

        WHERE { state } AND { ConstDatabaseCaseNaming.TID } = { paras[ConstDatabaseCaseNaming.TID] }

        ORDER BY ID DESC

        { limit }

        """

    @staticmethod
    def cases_add(paras: dict = None) -> str:
        """
        创建用例

        * 创建用例执行信息。

        参数:
            * tid (int): 所属任务记录ID，不可为空
            * name (str): 用例名称，不可为空
            * remark (str): 用例备注，不可为空

        """
        # 检查参数是否缺失
        paras = _machining_parameter(paras,
                                     [
                                         ConstDatabaseCaseNaming.TID,
                                         ConstDatabaseCaseNaming.NAME,
                                         ConstDatabaseCaseNaming.REMARK,
                                         ConstDatabaseCaseNaming.EXECUTE_DATE
                                     ])

        return f""" 

        INSERT INTO { ConstDatabaseTableNaming.CASES } (
        
            id, 
            { ConstDatabaseCaseNaming.TID }, 
            { ConstDatabaseCaseNaming.NAME }, 
            { ConstDatabaseCaseNaming.REMARK }, 
            { ConstDatabaseCaseNaming.STATE }, 
            { ConstDatabaseCaseNaming.EXECUTE_DATE }, 
            { ConstDatabaseCaseNaming.START_TIME }, 
            { ConstDatabaseCaseNaming.END_TIME },
            { ConstDatabaseCaseNaming.SPEND_TIME_S },
            { ConstDatabaseCaseNaming.RUN_STEP }
        )

        VALUES (
        
            NULL,
            '{ paras[ConstDatabaseCaseNaming.TID] }',
            '{ paras[ConstDatabaseCaseNaming.NAME] }',
            '{ paras[ConstDatabaseCaseNaming.REMARK] }',
            '{ ConstTaskAndCaseState.BLOCK }', 
            '{ paras[ConstDatabaseCaseNaming.EXECUTE_DATE] }',
             NULL, 
             NULL,
             NULL,
             NULL
        )

        """

    @staticmethod
    def cases_update(paras: dict = None) -> str:
        """
        更新用例

        * 更新用例信息。

        参数:
            * cid (int): 用例ID，不可为空
            * state (str): 用例状态，不可为空
            * start_time (date): 开始时间，不可为空
            * end_time (date): 结束时间，不可为空
            * spend_time_s (int): 消耗时间(秒)，不可为空
            * run_step (str): 运行步骤，不可为空

        """
        # 处理参数
        paras = _machining_parameter(paras,
                                     [
                                         ConstDatabaseCaseNaming.CID,
                                         ConstDatabaseCaseNaming.STATE,
                                         ConstDatabaseCaseNaming.START_TIME,
                                         ConstDatabaseCaseNaming.END_TIME,
                                         ConstDatabaseCaseNaming.SPEND_TIME_S,
                                         ConstDatabaseCaseNaming.RUN_STEP
                                     ])

        return f"""

        UPDATE { ConstDatabaseTableNaming.CASES }

        SET 
            { ConstDatabaseCaseNaming.STATE }         = '{ paras[ConstDatabaseCaseNaming.STATE] }',
            { ConstDatabaseCaseNaming.START_TIME }    = '{ paras[ConstDatabaseCaseNaming.START_TIME] }',
            { ConstDatabaseCaseNaming.END_TIME }      = '{ paras[ConstDatabaseCaseNaming.END_TIME] }',
            { ConstDatabaseCaseNaming.SPEND_TIME_S }  = '{ paras[ConstDatabaseCaseNaming.SPEND_TIME_S] }',
            { ConstDatabaseCaseNaming.RUN_STEP }      = '{ paras[ConstDatabaseCaseNaming.RUN_STEP] }'

        WHERE
            ID = { paras[ConstDatabaseCaseNaming.CID] }

        """

    @staticmethod
    def cases_delete(paras: dict = None) -> str:
        """
        删除用例

        * 删除指定用例ID的用例信息。

        参数:
            * cid (int): 用例ID，不可为空

        """
        # 检查参数是否缺失
        paras = _machining_parameter(paras, [ConstDatabaseCaseNaming.CID])

        return f"""

        DELETE FROM { ConstDatabaseTableNaming.CASES }

        WHERE ID = { paras[ConstDatabaseCaseNaming.CID] }

        """

    @staticmethod
    def cases_delete_all(paras: dict = None) -> str:
        """
        删除所有用例

        * 删除指定任务ID的所有用例信息。

        参数:
            * tid (int): 所属任务ID，不可为空

        """
        # 检查参数是否缺失
        paras = _machining_parameter(paras, [ConstDatabaseTaskNaming.TID])

        return f"""

        DELETE FROM { ConstDatabaseTableNaming.CASES }

        WHERE TID = { paras[ConstDatabaseTaskNaming.TID] }

        """


def _get_desc_limit_section(paras: dict):
    """
    获取数据返回量限制

    * 默认是获取 1 页， 10 条数据。

    """
    if 'page' not in paras.keys():
        paras['page'] = 1
    if 'number' not in paras.keys():
        paras['number'] = 10
    paras['page'] = int(paras['page'])
    paras['number'] = int(paras['number'])

    limit_begin = 0
    if paras['page'] > 1:
        limit_begin = (paras['page'] - 1) * paras['number']

    return f'LIMIT { limit_begin }, { paras["number"] }'


def _machining_parameter(paras: dict, essential_list: list):
    """
    检查参数是否缺失
    """
    if not paras:
        paras = {}

    for essential_para in essential_list:
        if paras[essential_para] is None:
            raise ParametersAreMissingException(essential_para)

    return paras


class ParametersAreMissingException(Exception):
    """
    参数缺失错误类
    """
    def __init__(self, missing_para_key):
        Exception.__init__(self, f'SQL必要参数 { missing_para_key } 缺失!')



# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


from selenium.webdriver.common.by import By

from rains.common.log import Log

from rains.kit.web.engine import WebEngine
from rains.kit.web.packaging.plant_view import WebPlantView
from rains.kit.web.packaging.plant_element import WebPlantElement
from rains.kit.web.packaging.element_structure import WebElementStructure


BY = By


class WebPlant(object):
    """
    WEB工厂构建类

    * 工厂构建类提供了对浏览器视图与控件元素的封装。
    * 浏览器视图: 封装了浏览器本身的功能，例如访问URL、管理标签页、控制弹窗、上传文件等。
    * 控件元素：封装了对页面元素的操作功能，如控制鼠标、控制输入框、下拉框、判断元素是否存在等。

    """

    _log = Log()
    """ 日志服务 """

    _engine: WebEngine
    """ 引擎对象 """

    def __init__(self, engine: WebEngine):
        """
        初始化

        Args:
            engine (WebEngine): WEB引擎对象

        """
        self._engine = engine

    @property
    def engine(self):
        return self._engine

    @property
    def driver(self):
        return self._engine.driver

    @property
    def view(self) -> WebPlantView:
        """
        视图

        * 提供针对浏览器本身的操作函数，例如访问URL，最大化窗口，截图PNG等等。

        """
        return WebPlantView(self._engine)

    def element(self, page: str, name: str, by_key: str, by_value: str,
                anchor_by_key: str = None,
                anchor_by_value: str = None,
                anchor_location_id: int = None) -> WebPlantElement:
        """
        元素控件

        * 基于元素特征构建指向该元素的元素控件对象。
        * 元素控件对象封装了针对元素的一些操作元素，例如鼠标点击、键盘输入等等。

        Args:
            page (str): 页面名称
            name (str): 组件名称
            by_key (str): 元素定位策略
            by_value (str): 元素定位策略对应的值
            anchor_by_key (str): 锚点元素定位策略
            anchor_by_value (str): 锚点元素定位策略
            anchor_location_id (str): 锚点元素定位ID

        """
        element_structure = WebElementStructure(page, name, by_key, by_value,
                                                anchor_by_key, anchor_by_value, anchor_location_id)

        return WebPlantElement(self._engine, element_structure)



# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


import sqlite3

from threading import Lock

from rains.common.db.const import ConstDatabaseTaskNaming
from rains.common.db.const import ConstDatabaseCaseNaming
from rains.common.db.const import ConstDatabaseTableNaming

from rains.common.const import ConstPath
from rains.common.decorator import singleton_pattern


@singleton_pattern
class Database(object):
    """
    数据库

    * 基于 SQLite 数据库。

    """

    _conn: sqlite3.Connection = None
    """ SQLite数据库连接对象 """

    _lock: Lock = Lock()
    """ 数据库互斥锁 """

    def __init__(self):
        """
        初始化
        """
        try:
            # 创建当前项目的日志根目录
            if not ConstPath.DATA.is_dir():
                ConstPath.DATA.mkdir(parents=True, exist_ok=True)

            # 单例创建/连接数据库
            if not self._conn:
                self.connect()
                self._conn.commit()

        except BaseException as e:
            raise Exception(f'SQLite数据库初始化异常:: { e } ')

    def read(self, sql: str) -> list:
        """
        读取

        * 执行只读查询语句，该函数不会获取数据库互斥锁。

        Args:
            sql (str) : SQLite 查询语句

        Return:
            查询结果列表

        """
        try:
            # 创建数据库游标
            cur = self._conn.cursor()
            # 执行查询语句
            cur.execute(sql)
            # 获取查询结果
            data = cur.fetchall()
            # 注销数据库游标
            cur.close()

            return data

        except BaseException as e:
            raise Exception(f'SQLite数据库读取异常:: { e } ')

    def write(self, sql: str) -> int:
        """
        写入

        * 执行互斥写入语句，该函数执行时会获取数据库互斥锁，结束时释放。

        Args:
            sql (str) : SQLite 语句

        Return:
            写入数据后返回该条数据的自增ID

        """
        try:
            # 获取锁
            with self._lock:

                # 创建数据库游标
                cur = self._conn.cursor()
                # 执行查询语句
                cur.execute(sql)
                # 获取数据自增ID
                aid = cur.lastrowid
                # 注销数据库游标
                cur.close()

                return aid

        except BaseException as e:
            raise Exception(f'SQLite数据库写入异常:: { e } ')

    def connect(self):
        """
        数据库连接
        """
        # 创建数据库连接对象
        self._conn = sqlite3.connect(ConstPath.DB, check_same_thread=False)
        # 创建数据库游标
        Cur = self._conn.cursor()
        # 构建任务表
        Cur.execute(CreateTableSql.TASKS)
        # 构建用例表
        Cur.execute(CreateTableSql.CASES)
        # 注销数据库游标
        Cur.close()

    def commit(self):
        """
        事务保存
        """
        self._conn.commit()

    def rollback(self):
        """
        事务回滚
        """
        self._conn.rollback()

    def close(self):
        """
        关闭数据库连接
        """
        self._conn.close()


class CreateTableSql(object):
    """
    建表语句
    """

    TASKS = f"""
        
    CREATE TABLE IF NOT EXISTS { ConstDatabaseTableNaming.TASKS }
    (
        id                                        INTEGER PRIMARY KEY,
        { ConstDatabaseTaskNaming.NAME }          TEXT NOT NULL,
        { ConstDatabaseTaskNaming.REMARK }        TEXT NOT NULL,
        { ConstDatabaseTaskNaming.EXECUTE_DATE }  DATE NOT NULL,
        { ConstDatabaseTaskNaming.IS_COMPLETED }  BLOB NOT NULL,
        { ConstDatabaseTaskNaming.START_TIME }    DATE,
        { ConstDatabaseTaskNaming.END_TIME }      DATE,
        { ConstDatabaseTaskNaming.SPEND_TIME_S }  INTEGER,
        { ConstDatabaseTaskNaming.CASE_ALL }      INTEGER,
        { ConstDatabaseTaskNaming.CASE_PASS }     INTEGER,
        { ConstDatabaseTaskNaming.CASE_FAIL }     INTEGER
    );

    """

    """
    创建任务表

    * 记录任务的执行信息。

    表结构:
        * id (int): 记录编号，随数据量而自增，无需手动输入
        * name (str): 任务名称，不可为空
        * remark (str): 任务备注信息，不可为空
        * execute_date (date): 执行日期，不可为空
        * is_completed (bool): 是否已完成，不可为空
        * start_time (date): 开始时间
        * end_time (date): 结束时间
        * spend_time_s (date): 消耗时间(秒)
        * case_all (int): 所有用例计数
        * case_pass (int): 成功的用例计数
        * case_fail (int): 失败用例数计数

    """

    CASES = f"""

    CREATE TABLE IF NOT EXISTS { ConstDatabaseTableNaming.CASES }
    (
        id                                        INTEGER PRIMARY KEY,
        { ConstDatabaseCaseNaming.TID }           INT NOT NULL,
        { ConstDatabaseCaseNaming.NAME }          TEXT NOT NULL,
        { ConstDatabaseCaseNaming.REMARK }        TEXT NOT NULL,
        { ConstDatabaseCaseNaming.STATE }         TEXT NOT NULL,
        { ConstDatabaseCaseNaming.EXECUTE_DATE }  DATE NOT NULL,
        { ConstDatabaseCaseNaming.START_TIME }    DATE,
        { ConstDatabaseCaseNaming.END_TIME }      DATE,
        { ConstDatabaseCaseNaming.SPEND_TIME_S }  INTEGER,
        { ConstDatabaseCaseNaming.RUN_STEP }      TEXT
    );
        
    """

    """
    创建用例表

    * 记录用例的执行信息。
    
    表结构:
        * id (int): 记录编号，随数据量而自增，无需手动输入
        * tid (int): 所属的任务记录编号，不可为空
        * name (str): 用例名称，不可为空
        * remark (str): 用例备注信息，不可为空
        * state (str): 用例状态，不可为空
        * start_time (date): 开始时间
        * end_time (date): 结束时间
        * spend_time_s (date): 消耗时间(秒)
        * run_step (str): 运行步骤
            
    """



# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


class MiniElementStructure(object):
    """
    小程序元素结构体
    """

    page: str
    """ 页面名称 """

    name: str
    """ 组件名称 """

    selector: str
    """ 元素选择器字符串 """

    inner_text: str
    """ 控件内的文字 """

    text_contains: str
    """ 控件内的模糊文字 """

    value: str
    """ 控件的 value 值 """

    def __init__(self,
                 page: str,
                 name: str,
                 selector: str,
                 inner_text: str = None,
                 text_contains: str = None,
                 value: str = None):
        """
        初始化

        * selector 仅支持下列语法:
            * ID选择器: #the-id
            * class选择器（可以连续指定多个）: .a-class.another-class
            * 标签选择器: view
            * 子元素选择器: .the-parent > .the-child
            * 后代选择器: .the-ancestor .the-descendant
            * 跨自定义组件的后代选择器: custom-element1>>>.custom-element2>>>.the-descendant
            * 多选择器的并集：#a-node, .some-other-nodes
            * xpath：可以在真机调试的 wxml pannel 选择节点->右键->copy->copy full xpath获取，暂不支持[text()='xxx']这类xpath条件

        Args:
            page (str): 页面名称
            name (str): 组件名称
            selector (str): css选择器或以/或//开头的xpath
            inner_text (str): 通过控件内的文字识别控件, xpath暂不支持
            text_contains (str): 通过控件内的文字模糊匹配控件, xpath暂不支持
            value (str): 通过控件的 value 识别控件, xpath暂不支持

        """
        # 获取元素参数
        self.page = page
        self.name = name
        self.selector = selector
        self.inner_text = inner_text
        self.text_contains = text_contains
        self.value = value



# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


from abc import ABCMeta
from abc import abstractmethod


class ICore(metaclass=ABCMeta):
    """
    核心接口
    """

    @abstractmethod
    def start_task(self, task):
        """
        开始执行 task

        * 接收一个 task，并执行该 task。

        Args:
            task (object): 任务对象，可以是实现 ITask 接口的任务类，或者是符合JSON译文标准的字典

        """



# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


class BrowserType(object):
    """
    浏览器类型
    """

    CHROME = "BROWSER_TYPE_CHROME"
    """ 谷歌浏览器 """



# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


class WebElementStructure(object):
    """
    WEB元素结构体
    """

    page: str
    """ 页面名称 """

    name: str
    """ 组件名称 """

    by_key: str
    """ 元素定位策略 """

    by_value: str = ''
    """ 元素定位策略对应的值 """

    anchor_by_key: str
    """ 锚点元素定位策略 """

    anchor_by_value: str
    """ 锚点元素定位策略 """

    anchor_location_id: int
    """ 锚点元素定位ID """

    def __init__(self,
                 page: str,
                 name: str,
                 by_key: str,
                 by_value: str,
                 anchor_by_key: str = None,
                 anchor_by_value: str = None,
                 anchor_location_id: int = None):
        """
        初始化

        * 构建组件时，需要传递组件于页面上的定位方式(by_key)与相应的值(by_value)。
        * 当传递了非必须的锚点元素时，表示将在该锚点节点中定位元素。
        * 如果传递锚点元素(anchor)，则必须同时传递 anchor_by_key、anchor_by_value、anchor_location_id。

        Args:
            page (str): 页面名称
            name (str): 组件名称
            by_key (str): 元素定位策略
            by_value (str): 元素定位策略对应的值
            anchor_by_key (str): 锚点元素定位策略
            anchor_by_value (str): 锚点元素定位策略
            anchor_location_id (str): 锚点元素定位ID

        """
        # 获取元素参数
        self.page = page
        self.name = name
        self.by_key = by_key
        self.by_value = by_value
        self.anchor_by_key = anchor_by_key
        self.anchor_by_value = anchor_by_value
        self.anchor_location_id = anchor_location_id



# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


import time

from minium import Page
from minium import WXMinium

from Rains2.common.log import Log
from Rains2.kit.mini.mini_element_structure import MiniElementStructure


class MiniFindHandle(object):
    """
    [ 小程序元素定位器 ]

    * 基于 minium 原生定位函数的封装。

    """

    _log = Log()
    """ [ 日志服务 ] """

    _driver: WXMinium
    """ [ 原生驱动对象 ] """

    def __init__(self, base_driver: WXMinium):
        """
        [ 初始化 ]

        Args:
            base_driver (WXMinium): 小程序驱动对象

        """
        self._driver = base_driver

    # func boundary -----------------------------------------------------------------------------------------------

    def find(self,
             mini_element_structure: MiniElementStructure,
             wait: int = 20) -> list:
        """
        [ 元素定位 ]

        * 根据参数描述从当前页面上定位元素列表。

        Args:
            mini_element_structure (str): 小程序元素结构体
            wait (int): 查找元素所等待的显式等待时长

        Return:
            已获取定位的元素列表

        """

        try:

            time.sleep(0.2)

            # 获取当前页面对象
            page: Page = self._driver.app.get_current_page()

            return page.get_elements(
                selector=mini_element_structure.selector,
                max_timeout=wait,
                inner_text=mini_element_structure.inner_text,
                text_contains=mini_element_structure.text_contains,
                value=mini_element_structure.value,
                index=-1
            )

        except BaseException as e:
            raise Exception(f'定位元素时发生了异常:: { e }')



# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


import pathlib


class ConstFileNaming(object):
    """
    文件命名
    """

    LOGS = 'logs'
    """ 项目日志目录 """

    DATA = 'data'
    """ 项目数据目录 """

    DB = 'database.db'
    """ 数据库文件 """


class ConstPath(object):
    """
    路径
    """

    ROOT = pathlib.Path().cwd()
    """ 项目根 """

    LOGS = ROOT.joinpath(ConstFileNaming.LOGS)
    """ 日志目录 """

    DATA = ROOT.joinpath(ConstFileNaming.DATA)
    """ 数据目录 """

    DB = DATA.joinpath(ConstFileNaming.DB)
    """ 数据库文件 """


class ConstProjectType(object):
    """
    项目类型
    """

    WEB = 'PROJECT_TYPE_FROM_WEB'
    """ WEB """

    MINI = 'PROJECT_TYPE_FROM_MINI'
    """ 微信小程序 """


class ConstTaskType(object):
    """
    任务类型
    """

    CODE = 'TASK_TYPE_FROM_CODE'
    """ 代码译文 """

    JSON = 'TASK_TYPE_FROM_JSON'
    """ JSON译文 """


class ConstTaskAndCaseState(object):
    """
    任务与用例的状态
    """
    BLOCK = '堵塞'
    """ 堵塞 """

    ANOMALY = '异常'
    """ 异常 """

    SUCCESSFUL = '成功'
    """ 成功 """

    UNSUCCESSFUL = '失败'
    """ 失败 """


class ConstTaskOrderSignNaming(object):
    """
    任务流程指令命名
    """

    CLASS_STARTING = 'set_class_starting'
    """ 类级起点 """

    CLASS_ENDING = 'set_class_ending'
    """ 类级终点 """

    FUNCTION_STARTING = 'set_function_starting'
    """ 函数级起点 """

    FUNCTION_ENDING = 'set_function_ending'
    """ 函数级终点 """



# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


from flask import jsonify
from flask import request

from rains.common.db.sql import Sql
from rains.common.db.database import Database


# 数据库实例
db: Database = Database()


def get_request_parameters(*paras) -> dict:
    """
    获取请求参数

    * 解析请求接口时携带的参数。

    Args:
        paras (*): 需要解析的参数 Key, 允许传递多个值。

    Return:
        返回参数字典

    """
    try:
        request_handle = request.args
        if request.method == 'POST':
            request_handle = request.form

        return_para_dict = {}
        for para_key in paras:
            para_value = request_handle.get(para_key)
            return_para_dict.update({para_key: para_value})

        return return_para_dict

    except BaseException as e:
        raise Exception(f"接口解析参数时，发生异常:: { e }")


def successful(paras: dict or None) -> jsonify:
    """
    请求成功返回参数

    Args:
        paras (dict or None): 需要返回的数据字典

    Return:
        经过加工的数据结构

    """
    r_structure = {
        'code': 200,
        'message': '请求成功',
        'data': {}
    }

    if paras:
        r_structure.update({'data': paras})

    return jsonify(r_structure)


def unsuccessful(err_message: str) -> jsonify:
    """
    请求失败返回参数

    Args:
        err_message (str): 需要返回的错误信息

    Return:
        经过加工的数据结构

    """
    r_structure = {
        'code': 500,
        'message': '请求失败',
        'err': err_message
    }

    return jsonify(r_structure)



# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


from rains.common.const import ConstTaskType
from rains.common.const import ConstTaskOrderSignNaming


CASE_SPECIFICATION_NAME = 'CASE'


class TaskInstruction(object):
    """
    任务指令集

    * rains 允许通过多种方式编写自动化测试用例。
    * 为了使这些方式具备一定的兼容性，TaskInstruction类将进行中间指令集的转换。

    """

    task_name: str
    """ 任务名称 """

    task_type: str
    """ 任务类型 """

    task_class = None
    """ 任务类的类名 """

    _is_packaging: bool = False
    """ 是否完成组装 """

    _instruction_map_flow: dict
    """ 流程指令集 """

    _instruction_map_step: dict
    """ 步骤指令集 """

    def __init__(self):
        """
        初始化
        """

        # 初始化 流程指令集
        self._instruction_map_flow = {
            ConstTaskOrderSignNaming.CLASS_STARTING: None,
            ConstTaskOrderSignNaming.CLASS_ENDING: None,
            ConstTaskOrderSignNaming.FUNCTION_STARTING: None,
            ConstTaskOrderSignNaming.FUNCTION_ENDING: None,
        }

        # 初始化 步骤指令集
        self._instruction_map_step = {}

    @property
    def instruction_map_flow(self):
        return self._instruction_map_flow

    @property
    def instruction_map_step(self):
        return self._instruction_map_step

    def convert_from_code(self, task_class):
        """
        从 Task 类中转换

        * NULL

        Args:
            task_class (Object): 继承 ITask 接口的任务类

        """
        try:

            if not self._is_packaging:

                # 确认 指令集类型
                self.task_type = ConstTaskType.CODE
                # 获取 任务类的类名
                self.task_class = task_class
                # 读取 任务类的信息
                self.task_name = task_class.__name__

                # 捕获指令集
                for k, v in self.task_class.__dict__.items():

                    # 捕获 流程指令集
                    # 遍历任务类中是否存在与 流程指令命名 一致的函数
                    # 如果存在则将其储存到 流程指令集 中
                    if k in self._instruction_map_flow.keys() and callable(v):
                        self._instruction_map_flow[k] = v

                    # 捕获 步骤指令集
                    # 遍历任务类中函数命名是否以符合规范
                    # 如果存在则将其储存到 步骤指令集 中
                    if str(k[:4:]).upper() == CASE_SPECIFICATION_NAME and callable(v):
                        self._instruction_map_step[k] = v

                # 完成封装
                self._is_packaging = True

            return self

        except BaseException as e:
            raise Exception(f'从 Task类 中转换指令集时发生了异常:: { e }')



# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


import time
import win32con
import win32gui

from selenium.common.exceptions import InvalidArgumentException

from rains.common.log import Log
from rains.common.task.case_recorder import CaseRecorder

from rains.kit.web.engine import WebEngine


class WebPlantView(object):
    """
    WEB视图

    * 提供针对浏览器本身的操作函数。

    """

    _log = Log()
    """ 日志服务 """

    _engine: WebEngine
    """ 引擎对象 """

    _case_recorder: CaseRecorder = CaseRecorder()
    """ 用例记录器 """

    def __init__(self, engine: WebEngine):
        """
        [ 初始化 ]

        Args:
            engine (WebEngine): WEB引擎对象

        """
        self._engine = engine

    @property
    def get(self):
        """
        WEB视图'获取'相关函数
        """
        return _WebPlantViewMapGet(self._engine, self._log_out, self._err_out)

    @property
    def set(self):
        """
        WEB视图'设置'相关函数
        """
        return _WebPlantViewMapSet(self._engine, self._log_out, self._err_out)

    @property
    def page(self):
        """
        WEB视图'页面'相关函数
        """
        return _WebPlantViewMapPage(self._engine, self._log_out, self._err_out)

    @property
    def alert(self):
        """
        WEB视图'弹窗'相关函数
        """
        return _WebPlantViewMapAlert(self._engine, self._log_out, self._err_out)

    def to_url(self, url: str, log: bool = True):
        """
        跳转URL

        * 跳转至指定的URL。

        Args:
            url (str): URL
            log (bool): 日志开关

        """
        try:
            self._engine.driver.get(url)
            self._log_out(f'跳转URL[{ url }]', log)
            return self

        except InvalidArgumentException:
            self._err_out(f'跳转URL异常:: URL参数可能错误!')

        except BaseException as e:
            self._err_out(f'跳转URL异常:: { e }')

    def refresh(self, log: bool = True):
        """
        刷新页面

        * 当前标签页刷新页面。

        Args:
            log (bool): 日志开关

        """
        try:
            self._engine.driver.refresh()
            self._log_out(f'刷新页面', log)
            return self

        except BaseException as e:
            self._err_out(f'刷新页面异常:: { e }')

    def back(self, log: bool = True):
        """
        回退页面

        * 当前标签页回退页面。

        Args:
            log (bool): 日志开关

        """
        try:
            self._engine.driver.back()
            self._log_out(f'回退页面', log)
            return self

        except BaseException as e:
            self._err_out(f'回退页面异常:: { e }')

    def forward(self, log: bool = True):
        """
        前滚页面

        * 当前标签页前滚页面。

        Args:
            log (bool): 日志开关

        """
        try:
            self._engine.driver.forward()
            self._log_out(f'前滚页面', log)
            return self

        except BaseException as e:
            self._err_out(f'前滚页面异常::{ e }')

    def save_png(self, save_path: str, log: bool = True):
        """
        保存快照

        * 将当前浏览器截图保存至 save_path 中。

        Args:
            save_path (str): 保存 .png 的路径
            log (bool): 日志开关

        """
        try:
            self._engine.driver.get_screenshot_as_file(str(save_path))
            self._log_out(f'保存快照[{save_path}]', log)
            return self

        except BaseException as e:
            self._err_out(f'保存快照异常::{ e }')

    def upload_local_file(self, file_path: str, log: bool = True):
        """
        上传本地文件

        Args:
            file_path (str): 本地文件路径
            log (bool): 日志开关

        """
        try:
            # 打开 upload 对话框需要耗时
            time.sleep(1.5)

            # 获取 upload 对话框
            dialog = win32gui.FindWindow('#32770', u'打开')

            # 定位路径输入框与确认按钮的对象句柄
            combobox_ex32 = win32gui.FindWindowEx(dialog, 0, 'ComboBoxEx32', None)
            combobox = win32gui.FindWindowEx(combobox_ex32, 0, 'ComboBox', None)
            edit = win32gui.FindWindowEx(combobox, 0, 'Edit', None)
            confirm_button = win32gui.FindWindowEx(dialog, 0, 'Button', None)

            # 调用句柄
            win32gui.SendMessage(edit, win32con.WM_SETTEXT, None, file_path)
            win32gui.SendMessage(dialog, win32con.WM_COMMAND, 1, confirm_button)

            self._log_out(f'上传本地文件', log)
            return self

        except BaseException as e:
            self._err_out(f'上传本地文件异常:: { e }')

    def _log_out(self, message: str, log: bool):
        """
        日志输出

        Args:
            message (str): 输出信息

        """
        if log:
            message = f'执行核心[{ self._engine.core_id }]::{ message }'
            self._log.info(message)
            self._case_recorder.write(self._engine.core_id, message)

        return message

    def _err_out(self, message: str):
        """
        错误输出

        Args:
            message (str): 输出信息

        """
        message = self._log_out(message, True)
        raise Exception(message)


class _WebPlantViewMapBase(object):
    """
    WEB视图封装基类
    """

    _web_engine: WebEngine
    """ 引擎对象 """

    _log_out = None
    """ 日志输出函数 """

    _err_out = None
    """ 错误输出函数 """

    def __init__(self, web_engine: WebEngine, log_output_function, err_output_function):
        """
        [ 初始化 ]

        Args:
            web_engine (WebEngine): WEB引擎对象
            log_output_function (): 日志输出函数
            err_output_function (): 错误输出函数

        """
        self._web_engine = web_engine
        self._log_out = log_output_function
        self._err_out = err_output_function


class _WebPlantViewMapGet(_WebPlantViewMapBase):
    """
    WEB视图'获取'相关函数
    """

    def url(self, log: bool = True) -> str:
        """
        获取当前标签页的URL

        Args:
            log (bool): 日志开关

        """
        try:
            current_url = self._web_engine.driver.current_url
            self._log_out(f'获取当前标签页的URL[{ current_url }]', log)
            return current_url

        except BaseException as e:
            self._err_out(f'获取当前标签页的URL异常:: { e }')

    def title(self, log: bool = True) -> str:
        """
        获取当前标签页的 Title

        Args:
            log (bool): 日志开关

        """
        try:
            current_title = self._web_engine.driver.title
            self._log_out(f'当前标签页的 Title [{ current_title }]', log)
            return current_title

        except BaseException as e:
            self._err_out(f'当前标签页的 Title 异常:: { e }')

    def window_list(self, log: bool = True):
        """
        获取标签页句柄列表

        Args:
            log (bool): 日志开关

        """
        try:
            window_handles = self._web_engine.driver.window_handles
            self._log_out(f'获取标签页句柄列表', log)
            return window_handles

        except BaseException as e:
            self._err_out(f'获取标签页句柄列表异常:: { e }')

    def window_current(self, log: bool = True):
        """
        获取当前标签页句柄

        Args:
            log (bool): 日志开关

        """
        try:
            current_window_handle = self._web_engine.driver.current_window_handle
            self._log_out(f'获取当前标签页句柄[{ current_window_handle }]', log)
            return current_window_handle

        except BaseException as e:
            self._err_out(f'获取当前标签页句柄异常:: { e }')

    def browser_size(self, log: bool = True) -> dict:
        """
        获取浏览器尺寸

        Args:
            log (bool): 日志开关

        Return:
            标识浏览器尺寸信息的字典

        """
        try:
            browser_size = self._web_engine.driver.get_window_size()
            self._log_out(f'获取浏览器尺寸[{ browser_size }]', log)
            return browser_size

        except BaseException as e:
            self._err_out(f'获取浏览器尺寸异常:: { e }')

    def browser_position(self, log: bool = True) -> dict:
        """
        获取浏览器位置坐标

        Args:
            log (bool): 日志开关

        Return:
            标识浏览器位置坐标信息的字典

        """
        try:
            browser_position = self._web_engine.driver.get_window_position()
            self._log_out(f'获取浏览器位置坐标[{ browser_position }]', log)
            return browser_position

        except BaseException as e:
            self._err_out(f'获取浏览器位置坐标异常:: { e }')

    def cookies(self, log: bool = True) -> list:
        """
        获取 cookies

        Args:
            log (bool): 日志开关

        Return:
            cookies 列表

        """
        try:
            cookies = self._web_engine.driver.get_cookies()
            self._log_out(f'获取cookies', log)
            return cookies

        except BaseException as e:
            self._err_out(f'获取cookies异常:: { e }')


class _WebPlantViewMapSet(_WebPlantViewMapBase):
    """
    WEB视图'设置'相关函数
    """

    def browser_size(self, width: int, height: int, log: bool = True):
        """
        设置浏览器尺寸

        Args:
            width (int): 宽度
            height (int): 高度
            log (bool): 日志开关

        """
        try:
            self._web_engine.driver.set_window_size(width, height)
            self._log_out(f'设置浏览器尺寸[{ width }::{ height }]', log)
            return self

        except BaseException as e:
            self._err_out(f'设置浏览器尺寸异常:: { e }')

    def browser_position(self, x: int, y: int, log: bool = True):
        """
        设置浏览器位置坐标

        Args:
            x (int): x坐标
            y (int): y坐标
            log (bool): 日志开关

        """
        try:
            self._web_engine.driver.set_window_position(x, y)
            self._log_out(f'设置浏览器位置坐标[{ x }::{ y }]', log)
            return self

        except BaseException as e:
            self._err_out(f'设置浏览器位置坐标异常:: { e }')

    def browser_max(self, log: bool = True):
        """
        设置浏览器窗口最大化

        Args:
            log (bool): 日志开关

        """
        try:
            self._web_engine.driver.maximize_window()
            self._log_out(f'设置浏览器窗口最大化', log)
            return self

        except BaseException as e:
            self._err_out(f'设置浏览器窗口最大化异常:: { e }')

    def browser_min(self, log: bool = True):
        """
        设置浏览器窗口最小化

        Args:
            log (bool): 日志开关

        """
        try:
            self._web_engine.driver.minimize_window()
            self._log_out(f'设置浏览器窗口最小化', log)
            return self

        except BaseException as e:
            self._err_out(f'设置浏览器窗口最小化异常:: { e }')

    def del_cookies(self, log: bool = True):
        """
        删除 cookies

        Args:
            log (bool): 日志开关

        """
        try:
            self._web_engine.driver.delete_all_cookies()
            self._log_out(f'删除cookies', log)
            return self

        except BaseException as e:
            self._err_out(f'删除cookies异常:: { e }')

    def add_cookies(self, cookie_list: list, log: bool = True):
        """
        添加 cookies

        Args:
            cookie_list (list): cookies 列表
            log (bool): 日志开关

        """
        try:
            for i in cookie_list:
                self._web_engine.driver.add_cookie(i)

            self._log_out(f'设置cookies', log)
            return self

        except BaseException as e:
            self._err_out(f'设置cookies异常:: { e }')


class _WebPlantViewMapPage(_WebPlantViewMapBase):
    """
    WEB视图'页面'相关函数
    """

    def open(self, url: str = '', log: bool = True):
        """
        打开新的标签页

        * 在当前浏览器中, 打开一个新的标签页。

        Args:
            url (str): 新标签页的URL，默认为空
            log (bool): 日志开关

        """
        try:
            # 通过JS语句打开新标签页
            self._web_engine.driver.execute_script(f'window.open("{url}")')
            # 将当前句柄跳转至新标签页
            self._web_engine.driver.switch_to.window(self._web_engine.driver.window_handles[-1])
            self._log_out(f'打开新标签页[{ len(self._web_engine.driver.window_handles) - 1 }]->[{ url }]', log)
            return self

        except BaseException as e:
            self._err_out(f'打开新标签页异常:: { e }')

    def switch(self, number: int, log: bool = True):
        """
        切换至标签页

        * 在当前浏览器中, 切换至编号为 [number] 的已存在标签页。

        Args:
            number (str): 标签页编号
            log (bool): 日志开关

        """
        try:
            self._web_engine.driver.switch_to.window(self._web_engine.driver.window_handles[number])
            self._log_out(f'切换至标签页[{ number }]', log)
            return self

        except IndexError:
            self._err_out(f'切换至标签页异常:: 没有这样的标签页，请重新确认标签页编号！')

        except BaseException as e:
            self._err_out(f'切换至标签页异常:: { e }')

    def close(self, number: int, log: bool = True):
        """
        关闭标签页

        * 在当前浏览器中, 关闭编号为 [number] 的已存在标签页。

        Args:
            number (str): 标签页编号
            log (bool): 日志开关

        """
        try:
            if len(self._web_engine.driver.window_handles) > 1:
                self.switch(number, log=False)
                self._web_engine.driver.close()
                self._web_engine.driver.switch_to.window(self._web_engine.driver.window_handles[-1])
                self._log_out(f'关闭标签页[{ number }]', log)
                return self

        except IndexError:
            self._err_out(f'切换至标签页异常:: 没有这样的标签页，请重新确认标签页编号！')

        except BaseException as e:
            self._err_out(f'切换至标签页异常:: { e }')

    def close_all(self, log: bool = True):
        """
        关闭所有标签页

        * 在当前浏览器中, 关闭所有已存在的标签页, 仅保留一个空白标签。

        Args:
            log (bool): 日志开关

        """
        try:
            self._web_engine.reset()
            self._log_out(f'关闭全部标签页', log)
            return self

        except BaseException as e:
            self._err_out(f'关闭全部标签页异常:: { e }')


class _WebPlantViewMapAlert(_WebPlantViewMapBase):
    """
    WEB视图'弹窗'相关函数
    """

    def get_text(self, log: bool = True) -> str:
        """
        获取弹窗文本

        Args:
            log (bool): 日志开关

        Return:
            弹窗文本字符串

        """
        try:
            alert_text = self._web_engine.driver.switch_to.alert.text
            self._log_out(f'获取弹窗文本[{ alert_text }]', log)
            return alert_text

        except BaseException as e:
            self._err_out(f'获取弹窗文本异常:: { e }')

    def tap_accept(self, log: bool = True):
        """
        点击弹窗确定按钮

        Args:
            log (bool): 日志开关

        """
        try:
            self._web_engine.driver.switch_to.alert.accept()
            self._log_out(f'点击弹窗确定按钮', log)
            return self

        except BaseException as e:
            self._err_out(f'点击弹窗确定按钮异常:: { e }')

    def tap_dismiss(self, log: bool = True):
        """
        点击弹窗取消按钮

        Args:
            log (bool): 日志开关

        """
        try:
            self._web_engine.driver.switch_to.alert.dismiss()
            self._log_out(f'点击弹窗取消按钮', log)
            return self

        except BaseException as e:
            self._err_out(f'点击弹窗取消按钮异常:: { e }')

    def send_key(self, key: str, log: bool = True):
        """
        弹窗输入文本

        Args:
            key (str): 文本
            log (bool): 日志开关

        """
        try:
            self._web_engine.driver.switch_to.alert.send_keys(key)
            self._log_out(f'弹窗输入文本[{ key }]', log)
            return self

        except BaseException as e:
            self._err_out(f'弹窗输入文本异常:: { e }')



# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


from rains.kit.web.plant import WebPlant


class WebModel(object):
    """
    WEB模型
    """

    plant: WebPlant

    def __init__(self, web_plant: WebPlant):
        """
        初始化

        * 构建页面模型主要由 Plant 构建类完成，实例化页面模型时需传递 Plant 实例。

        """
        self.plant = web_plant



import setuptools


with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()


setuptools.setup(

    name='rains',
    version='0.0.5',
    author='7cat',
    author_email='quinn.7@foxmail.com',
    description='NULL',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitee.com/catcat7/rains',
    packages=setuptools.find_packages(),

    install_requires=[
        'Selenium == 3.14.1',
        'pypiwin32',
        'psutil',
    ],

    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


import math

from flask import Flask
from flask import render_template

from rains.common.api.common import *

from rains.common.api.blueprint.data import data
from rains.common.api.blueprint.task import task
from rains.common.api.blueprint.case import case

from rains.common.db.const import ConstDatabaseTaskNaming
from rains.common.db.const import ConstDatabaseCaseNaming


app = Flask(__name__)
app.register_blueprint(data)
app.register_blueprint(task)
app.register_blueprint(case)
app.config['JSON_AS_ASCII'] = False
app._static_folder = './static'


@app.route('/', methods=['GET'])
@app.route('/index', methods=['GET'])
def index():
    """
    首页
    """
    page_data = {}

    new_exec_data = db.read(Sql.tasks_get_date_list())[0][0]
    page_data.update({'PageName': 'Index'})
    page_data.update({'NewExecData': new_exec_data})

    page_data.update({'NewTaskCount': db.read(Sql.tasks_get_count_from_data({
        ConstDatabaseTaskNaming.EXECUTE_DATE: new_exec_data
    }))[0][0]})
    page_data.update({'NewCaseCount': db.read(Sql.cases_get_count_form_data({
        ConstDatabaseCaseNaming.EXECUTE_DATE: new_exec_data
    }))[0][0]})
    page_data.update({'NewCaseFailCount': db.read(Sql.cases_get_all_count_in_fail_from_data({
        ConstDatabaseTaskNaming.EXECUTE_DATE: new_exec_data
    }))[0][0]})
    page_data.update({'NewSpendTime': round((db.read(Sql.tasks_get_spend_time_from_data({
        ConstDatabaseTaskNaming.EXECUTE_DATE: new_exec_data
    }))[0][0] / 60), 2)})

    page_data.update({'HistoryTaskCount': db.read(Sql.tasks_get_count())[0][0]})
    page_data.update({'HistoryCaseCount': db.read(Sql.cases_get_all_count())[0][0]})
    page_data.update({'HistoryCaseFailCount': db.read(Sql.cases_get_all_count_in_fail())[0][0]})
    page_data.update({'HistorySpendTime': round((db.read(Sql.tasks_get_spend_time())[0][0] / 60), 2)})
    page_data.update({'FailTaskCount': db.read(Sql.tasks_get_count_in_fail())[0][0]})

    return render_template('index.html', Data=page_data)


@app.route('/tasks/<page>', methods=['GET'])
def tasks(page):
    """
    任务
    """
    page = int(page)
    page_data = {}

    page_data.update({'PageName': 'Tasks'})
    page_data.update({'TaskList': db.read(Sql.tasks_get_all_item({'page': page}))})

    all_page = math.ceil(db.read(Sql.tasks_get_count())[0][0] / 10)
    next_page = page + 1 if page < all_page else page
    base_page = page - 1 if page > 1 else 1

    page_data.update({'CurrentPage': page})
    page_data.update({'AllPage': all_page})
    page_data.update({'NextPage': next_page})
    page_data.update({'BasePage': base_page})

    return render_template('tasks.html', Data=page_data)


@app.route('/cases/<tid>', methods=['GET'])
def cases(tid):
    """
    用例
    """
    page_data = {}
    page_data.update({'PageName': 'Cases'})

    # 转换用例格式
    base_case_list = db.read(Sql.cases_get_all_item({'tid': tid}))
    new_case_list = []
    for case_ in base_case_list:
        new_case_info = []
        number = 0
        for c_i in case_:
            if number == 9:
                c_i = c_i.split('\n')
            new_case_info.append(c_i)
            number += 1
        new_case_list.append(new_case_info)
    page_data.update({'CaseList': new_case_list})

    return render_template('cases.html', Data=page_data)



# !/usr/bin/env python
# coding:utf-8

# Copyright (c) 2021. quinn.7@foxmail.com All rights reserved.

# Based on the Apache License 2.0 open source protocol.

__author__ = 'quinn.7@foxmail.com'


from selenium.common.exceptions import TimeoutException
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support import expected_conditions as ec
from selenium.webdriver.chrome.webdriver import RemoteWebDriver

from rains.common.log import Log
from rains.kit.web.packaging.element_structure import WebElementStructure


class WebFindHandle(object):
    """
    WEB元素定位器
    """

    _log = Log()
    """ 日志服务 """

    _driver: RemoteWebDriver
    """ 原生驱动对象 """

    _web_driver_wait: WebDriverWait = WebDriverWait
    """ 原生驱动显示等待封装 """

    def __init__(self, base_driver: RemoteWebDriver):
        """
        初始化

        Args:
            base_driver (RemoteWebDriver): 原生驱动对象

        """
        self._driver = base_driver

    def find(self, element_structure: WebElementStructure,
             wait: int = 15, tolerance: bool = False) -> list:
        """
        元素定位

        * 接受一个 WebElementStructure 对象，并根据其描述从当前页面上定位元素列表。

        Args:
            element_structure (ElementStructure): 元素结构体
            wait (int): 查找元素所等待的显式等待时长
            tolerance (bool): 容错模式，为 True 则表示定位不到元素也不报错，返回空列表

        Return:
            已获取定位的元素列表

        """
        # 解析元素搜索范围
        search_range = self._analysis_search_range(element_structure, wait)

        # 开始定位UI元素
        elements = self._base_explicitly_wait_find_element(
            by_key=element_structure.by_key,
            by_value=element_structure.by_value,
            driver=search_range,
            wait=wait,
            tolerance=tolerance)

        return elements

    def _analysis_search_range(self, element_structure: WebElementStructure, wait: int):
        """
        解析元素搜索范围

        * 如果元素结构体中携带了锚元素则会返回定位的锚元素对象, 否则返回驱动。

        Args:
            element_structure (ElementStructure): WEB元素结构体
            wait (int): 查找元素所等待的显式等待时长

        Return:
            锚元素对象 or 驱动对象

        """
        search_range = self._driver

        if element_structure.anchor_by_key:
            search_range = self._base_explicitly_wait_find_element(
                by_key=element_structure.anchor_by_key,
                by_value=element_structure.anchor_by_value,
                driver=search_range,
                wait=wait,
                tolerance=False)[element_structure.anchor_location_id]

        return search_range

    def _base_explicitly_wait_find_element(self,
                                           by_key: str,
                                           by_value: str,
                                           driver,
                                           wait: int,
                                           tolerance: bool) -> list:
        """
        原生 WebDriverWait 封装

        * 封装 WebDriverWait 全局显示等待获取函数。
        * 该函数将在每 0.5 秒时遍历一次页面，直到找到元素或者等待(wait)超时。
        * 如果容错开关(tolerance)为 True, 则表示定位不到元素时返回空列表。

        Return:
            已获取定位的元素列表 or 空列表

        """
        try:
            return self._web_driver_wait(driver, wait, poll_frequency=0.5, ignored_exceptions=None)\
                .until(ec.presence_of_all_elements_located((by_key, by_value)))

        except TimeoutException:
            # 触发异常时，如果容错开关(tolerance)为 True，则不抛出异常，仅返回空列表
            if tolerance:
                return []
            else:
                raise Exception(f'定位器获取不到元素！')
